/* Objects.c
 *
 * This file contains routines that are responsible for handling object
 * communication.
 *
 */
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <netdb.h>
#include <stdio.h>
#include <signal.h>

#include "select.h"
#include "List.h"
#include "sockets.h"
#include "network.h"
#include "spede.h"
#include "objects.h"
#include "misc.h"
#include "spede_io.h"
#include "dean_objs.h"


#define TRUE 1
#define FALSE 0


/*****************************************************************************
 *                                                                           *
                                 Global Variables                            
 *                                                                           *
 *****************************************************************************/

int next_object_id, max_objects;
List objects;                          /* List of current objects */
List new_o_queue;                      /* New objects are put in here */
List connect_list;                   /* List of which apps are connected     */
extern int errno;
unsigned char connect_id = 1;
extern node_des local_node;


/*****************************************************************************
 *                                                                           *
                                   Functions 
 *                                                                           *
 *****************************************************************************/

/* INIT_OBJECTS
 *
 * This routine is called by dean main to initialize any variables in dean_objs
 */
int init_objects()
{
  objects = CreateList();
  new_o_queue = CreateList();
  connect_list = CreateList();
  return(0);
}


/* LAUNCH_OBJECT
 * 
 * This routine attempts to launch the named object. The path for the
 * compiled object is compiled in from the Makefile in a define called
 * O_PATH. If the object does not exist, the routine returns -1.
 */
int launch_object(object_name)
     char *object_name;
{
  char comm[256];
  char tempstr[256], *path;
  extern char *base_dir;
  extern char reghost[];
  extern int reg_port;
  char **argv;

  argv = (char **) malloc(6 * sizeof(char *));
  argv[0] = SaveString(object_name);
  argv[1] = SaveString("-h");
  printf("Here reghost is %s\n", reghost);
  argv[2] = SaveString(reghost);
  argv[3] = SaveString("-p");
  printf("Here\n");
  sprintf(tempstr, "%d", reg_port);
  argv[4] = SaveString(tempstr);
  argv[5] = 0;
  
  path = (char *) getenv("HOME");
  if (strcmp(path, ""))
    strcat(path, "/");

  if (!fork()) {
    sprintf(comm, "%s%s/%s", path, base_dir, object_name);
    
    if (execve(comm, argv, 0) == -1) {
      log("DEAN - Error while trying to launch object '%s'.\n", comm);
      return(-1);
    }
  }
  return(0);
}


/* NEW_CONNECTION
 *
 * This routine is called when a new connection is made to the dean, or when
 * the dean makes a connection with another dean.
 * A connection data entry is constructed to keep track of who is using the
 * link.
 */
void new_connection(ce, type)
     connect_entry *ce;
     int type;
{
  connect_des *new;

  new = (connect_des *) malloc(sizeof(connect_des));
  new->type = OBJECT;
  new->num_links = 1;
  new->ce = ce;
  new->connect_id = connect_id;
  new->type = type;
  connect_id = (connect_id+1) % 256;
  copy_node_des(&new->node, &ce->node);

  ce->data = new;
  
  /* OBJECT entries go at the beginning of the List, DEAN at the end */
  if (type == OBJECT)
    SetBeginning(connect_list);
  else
    SetEnd(connect_list);

  new->nodeptr = InsertData(connect_list, 0, (void *) new);
}


/* INSERT_OBJECT_LIST
 * 
 * This routine inserts an object into our list of objects. The routine
 * returns a pointer to a object_ent structure, which we use as the object
 * reference number.
 */
long insert_object_list(object, app_ref, app_socket, port_num)
     object_instance *object;
     long app_ref;
     int app_socket;
     long port_num;
{
  object_ent *new_ent;
  
  new_ent = (object_ent *) malloc(sizeof(object_ent));
  new_ent->object = object;
  new_ent->debug = object->debug;
  new_ent->link_ref = app_ref;
  new_ent->link_socket = app_socket;
  new_ent->is_link = FALSE;
  new_ent->port_num = port_num;
  log("DEAN - Object loaded on port %d\n", port_num);
  SetBeginning(objects);
  new_ent->nodeptr = InsertData(objects, 0, new_ent);

  object->data = (void *) new_ent;
  return((long) new_ent);
}


void queue_object(comm)
     command *comm;
{
  command *new;

  new = (command *) malloc(sizeof(command));
  *new = *comm;
  SetEnd(new_o_queue);
  InsertData(new_o_queue, 0, new);
}


/* LOG_COMMAND
 *
 * This routine is called when a message should be logged. The routine
 * assumes that a link has already been established with the correct object. 
 */
void log_command(comm, ent)
     command *comm;
     object_ent *ent;
{
  command s_comm;

  if (ent->debug) {
    log("DEAN - logging message to socket %d\n", ent->debug_link->ce->the_socket);
    s_comm.comm = LOG_MESSAGE;
    /* REF NUM ???? */
    s_comm.data.log.comm = comm;
    write_command(&s_comm, ent->debug_link->ce->the_socket);
  }
}


/* CONNECT_REMOTE_DEAN
 *
 * This routine attempts to connect with a remote dean. This call is used to
 * establish a link with another dean.
 */
connect_entry *connect_remote_dean(node)
     node_des *node;
{
  extern connect_type *connections;
  int the_socket;
  connect_entry *ce;
  short running;
  long port_num;

  query_server(node->name, "dean", &running, &port_num);
  the_socket = connect_remote(port_num, node->addr, 0);
  if (the_socket == 0) {
    log("DEAN - Error trying to connect to dean at addr %X\n", node->addr);
    return(0);
  }

  ce = add_connection(connections, the_socket, node);
  open_commands(the_socket);
  return(ce);
}


a_link *link_object(node)
     link_des *node;
{
  connect_des *cd;
  a_link *new_link;
  command comm;
  connect_entry *dean_ent;
  
  new_link = (a_link *) malloc(sizeof(a_link));

  log("DEAN - Linking to object ref %d\n", node->obj_ref);

  /* Check to see if link exists */
  SetBeginning(connect_list);
  while(!EndList(connect_list)) {
    ReadData(connect_list, 0, (void **) &cd);
/*    log("DEAN - Checking link to object id %d\n", cd->obj_ref); */
    if (node->node.addr == cd->node.addr) {            /* Found remote node */
      if (cd->type == OBJECT && (node->obj_ref == cd->obj_ref)) {
	log("DEAN - Found a link to use...\n");
	cd->num_links++;
	new_link->ce = cd->ce;
	new_link->link_ref = node->obj_ref;
	return(new_link);
      }
      if (cd->type == DEAN) {
	build_LINK_OBJECT(&comm, node->obj_ref, 0);
	write_command(&comm, cd->ce->the_socket);
	new_link->ce = cd->ce;
	new_link->link_ref = node->obj_ref;
	return(new_link);
      }
    }
    SetNext(connect_list);
  }
  
  /* NO LINK EXISTS, we need to make one */

  if (node->node.addr == local_node.addr) {
    log("DEAN - ERROR!! could not find a link to an object...EXITING\n");
    exit(1);
  }

  log("DEAN - no local links to that object, creating a link\n");
  dean_ent = connect_remote_dean(&node->node);
  if (dean_ent == 0) {
    log("Link to dean failed\n");
    return(0);
  }
    
  new_connection(dean_ent, DEAN);

  build_LINK_OBJECT(&comm, node->obj_ref, 0);
  write_command(&comm, dean_ent->the_socket);
  new_link->ce = dean_ent;
  new_link->link_ref = node->obj_ref;
  return(new_link);
}




/*****************************************************************************
 *                                                                           *
 *                            Protocol Functions                             *
 *                                                                           *
 *****************************************************************************/



int f_new_object(comm)
     command *comm;
{
  object_instance *object;

  object = &comm->data.new_o.object;

  launch_object(object->type);
  queue_object(comm);                      /* Wait until the object comes up */
}



int f_object_created(comm, ce)
     command *comm;
     connect_entry *ce;
{
  int socket;
  object_ent *ent;

  ent = (object_ent *) comm->recv_ref;

  ent->obj_ref = comm->send_ref;
  log("DEAN - (%s) object_ref is %d, dean_ref is %d \n", ent->object->type, ent->obj_ref, (long) ent);
  ent->obj_connect = ce;

  /* The connection now knows what object it is connected with */
  ((connect_des *) ce->data)->obj_ref = comm->recv_ref;

  socket = ent->link_socket;
  comm->send_ref = comm->recv_ref;
  comm->recv_ref = ent->link_ref;
  
/*  log("Rerouting OBJECT_CREATED to socket %d\n", socket); */
  write_command(comm, socket);

  if (ent->debug) {
    log("DEAN - OBJECT should be debugged...linking to debug site.\n");
    ent->debug_link = link_object(&ent->object->creator);
  }
}



int f_delete_object(comm)
     command *comm;
{
  object_ent *ent;
  
  ent = (object_ent *) comm->recv_ref;
  comm->recv_ref = ent->obj_ref;
  comm->send_ref = (long) ent;
  write_command(comm, ent->obj_connect->the_socket);
}


int f_start_time(comm)
     command *comm;
{
  object_ent *ent;
  
  ent = (object_ent *) comm->recv_ref;
  comm->recv_ref = ent->obj_ref;
  comm->send_ref = (long) ent;                         /* This is not used */
  write_command(comm, ent->obj_connect->the_socket);
}


int f_get_time(comm)
     command *comm;
{
  object_ent *ent;
  
  ent = (object_ent *) comm->recv_ref;
  comm->recv_ref = ent->obj_ref;
  comm->send_ref = (long) ent;
  write_command(comm, ent->obj_connect->the_socket);
}


int f_send_times(comm)
     command *comm;
{
  object_ent *ent;
  
  ent = (object_ent *) comm->recv_ref;
  comm->recv_ref = ent->link_ref;
  write_command(comm, ent->link_socket);
}


int delete_all_objects(socket)
     int socket;
{
  object_ent *ent;
  command comm;
  extern connect_type *connections;

  SetBeginning(objects);
  while(!EndList(objects)) {
    ReadData(objects, 0, (void **) &ent);
    if (ent->obj_connect->the_socket == socket) {
      delete_connection(connections, ent->obj_connect);
      if (!ent->is_link)
	free_object(ent->object);
      DeleteNode(objects);
      free(ent);
    }
    else if (ent->link_socket == socket) {
      log("DEAN - Terminating object id %d due to socket error\n", ent->object->id);
      ent->link_socket = -1;
      comm.recv_ref = ent->obj_ref;

      kill(ent->proc_id, SIGKILL);
/*      build_DELETE_OBJECT(&comm, 0, ent->obj_ref, (long) ent);
      write_command(&comm, ent->obj_connect->the_socket); */
    }
    SetNext(objects);
  }
}



int f_object_deleted(comm)
     command *comm;
{
  extern connect_type *connections;
  int r_socket;
  object_ent *ent;

  ent = (object_ent *) comm->recv_ref;

  comm->recv_ref = ent->link_ref;
  if (ent->link_socket != -1) {
    log("DEAN - Rerouting OBJECT_DELETED to socket %d\n", ent->link_socket);
    write_command(comm, ent->link_socket);
  }
  else
    log("DEAN - OBJECT_DELETED not forwarded...\n");

  
  delete_connection(connections, ent->obj_connect);
  if (!ent->is_link)
    free_object(ent->object);
  
  SetNode(objects, ent->nodeptr);
  DeleteNode(objects);
  free(ent);  
  return(0);
}

object_ent *create_the_object(comm, ce, port_num)
     command *comm;
     connect_entry *ce;
     long port_num;
{
  command s_comm;
  long ref;
  object_instance *object;
  object_ent *ent;

/*  log("Create Local Object\n"); */

  object = new_object();
  *object = comm->data.new_o.object;

  ref = insert_object_list(object, comm->send_ref, comm->socket, port_num);
  ent = (object_ent *) ref;

  copy_node_des(&object->location.node, &local_node);
  object->location.obj_ref = ref;

  comm->send_ref = ref;
  write_command(comm, ce->the_socket);
  return(ent);
}



/* SENT by object when object file is executed. File is either executed by
 * the dean, or from the command line
 */  
int f_object_load(comm, ce)
     command *comm;
     connect_entry *ce;
{
  command *s_comm;
  command obj_comm;
  long ref;
  object_instance *object;
  object_ent *ent;

  /* Check to see if object was loaded because we loaded it! */
  SetBeginning(new_o_queue);
  while(!EndList(new_o_queue)) {
    ReadData(new_o_queue, 0, (void **) &s_comm);
    if (!strcmp(comm->data.obj_load.object_name, 
		s_comm->data.new_o.object.type)) {
      ent = create_the_object(s_comm, ce, comm->data.obj_load.port_num);
      ent->proc_id = comm->data.obj_load.proc_id;
      log("DEAN - That object is on proc %d\n", ent->proc_id);
      free(s_comm);                          /* Don't free object memory */
      DeleteNode(new_o_queue);
      return;
    }
    SetNext(new_o_queue);
  }

  /* We didn't load the object, so the user must have launched it */
  
  log("DEAN - Object Free-launched on socket %d\n", comm->socket);
  object = new_object();
  ref = insert_object_list(object, comm->send_ref, comm->socket, comm->data.obj_load.port_num);
  ((object_ent *) object->data)->obj_connect = ce;

  /* Now the connection knows which object it is connect to */
  ((connect_des *) ce->data)->obj_ref = ref;

  build_OBJECT_LOADED(&obj_comm, ref);
  write_command(&obj_comm, comm->socket);
  
  log("Sent free launch command to socket %d\n", comm->socket);
/*  log("ADDR is %X, size of command %d\n", &comm->socket, sizeof(command)); */
}

 

int f_object_message(comm)
     command *comm;
{
  object_ent *ent;

  ent = (object_ent *) comm->recv_ref;

  comm->recv_ref = ent->obj_ref;
  comm->send_ref = (long) ent;
  printf("Writing command to object\n");
  write_command(comm, ent->obj_connect->the_socket);
  printf("Writing command to log\n");
  log_command(comm, ent);
  printf("Cleaning command\n");
  
  clean_command(comm);
  return(0);
}
  


int f_return_message(comm)
     command *comm;
{
  short r_socket;
  object_ent *ent;
  object_instance *object;

  ent = (object_ent *) comm->recv_ref;
  
  r_socket = ent->link_socket;
  comm->recv_ref = ent->link_ref;
  log("DEAN - Forwarding RETURN_MESSAGE to socket %d\n", r_socket);
  write_command(comm, r_socket);
  log_command(comm, ent);
  clean_command(comm);
  log("DEAN - Out of routine F_RETURN_MESSAGE\n");
  return(0);
}


int f_link_object(comm)
     command *comm;
{
  object_ent *ent, *new_ent;
  command s_comm;

  log("DEAN - request to link with object ref %d\n", comm->recv_ref);
  new_ent = (object_ent *) malloc(sizeof(object_ent));
  
  ent = (object_ent *) comm->recv_ref;
  build_OBJECT_LINKED(&s_comm, 0, (long) new_ent, ent->object);

  write_command(&s_comm, comm->socket);
  log("DEAN - changing link ref to %d\n", (long) new_ent);

  *new_ent = *ent;
  new_ent->link_socket = comm->socket;
  SetBeginning(objects);
  new_ent->nodeptr = InsertData(objects, 0, (void *) new_ent);
  new_ent->is_link = TRUE;
  new_ent->link_ref = comm->send_ref;
}


int f_speed_link(comm)
     command *comm;
{
  object_ent *ent;
  command s_comm;

  ent = (object_ent *) comm->recv_ref;
  build_SPEED_MADE(&s_comm, 0, ent->obj_ref, ent->port_num);
  write_command(&s_comm, comm->socket);
}
  
  
