/* Misc.c
 *
 * This file contains miscellaneous routines.
 */
#include <string.h>
#include <stdio.h>
#include "List.h"
#include <sys/time.h>
#include <ctype.h>

#define TRUE  1
#define FALSE 0

/*****************************************************************************
 *                                                                           *
 *                            Public Functions                               *
 *                                                                           *
 *****************************************************************************/

/* NUM_O_ITEMS
 *
 * This function returns the number of 'items' in a character string. An 'item'
 * is defined as a string of characters, or NULL, followed by a character.
 *
 * Examples: (where separator is a comma)
 *    "jim, bob, fred" returns 3 ("jim", "bob", "fred")
 *    ",,"             returns 3 ("", "", "")
 *    "jim,"           returns 2 ("jim", "")
 *    ""               returns 1 ("")
 */
int num_o_items(c, s)
     char c;
     char *s;
{
  register int num_items = 1;

  while (*s != 0) {
    if (*s == c)
      num_items++;
    s++;
  }
  return(num_items);
}


int num_items(s)
     char *s;
{
  return(num_o_items(',', s));
}


char *ptr_of_o_item(c, s, num)
     char c;
     char *s;
     int num;
{
  num--;
  while((num>0) && *s != 0) {             /* Move pointer to given item */
    if (*s == c)                          /* Comma means new item coming up */
      num--;
    s++;
  }

  while(isspace(*s) && *s != 0)            /* Advance start past white space */
    s++;

  return(s);
}


/* GET_ITEM
 *
 * This routine copies the text out of a given item. If that item does not
 * exist, NULL is copied into the string. White space before and after the
 * item are NOT copied.
 *
 * Examples:
 *
 *   "jim, bob  , fred" asking for item 2 returns "bob"
 *   ",,"               asking for item 2 returns NULL
 *  
 */
void get_o_item(c, s, dest, num)
     char c;
     char *s;
     char *dest;
     int num;
{
  char *d = dest;
  
  num--;
  while((num>0) && *s != 0) {             /* Move pointer to given item */
    if (*s == c) {                        /* means new item coming up */
      s++;
      while(isspace(*s) && *s != 0)
	s++;
      num--;
    }
    else
      s++;
  }

  if (*s == 0) {                           /* That item does not exist */
    *d = 0;
    return;
  }

  while(*s != c && *s != 0)                /* Copy until end of item */
    *d++ = *s++;

  *d = 0;

  d--;                                     /* Strip trailing white space */
  while(d >= dest && isspace(*d))
    *d-- = 0;
}


void get_item(s, dest, num)
     char *s;
     char *dest;
     int num;
{
  get_o_item(',', s, dest, num);
}


/* ALLOC_ITEM
 *
 * This routine returns a pointer to a 'malloced' string for the 'num'th item
 * in a string of characters.
 */
char *alloc_o_item(c, s, num)
     char c;
     char *s;
     int num;
{
  char temp[256], *r;
  
  get_o_item(c, s, temp, num);
  r = (char *) malloc(strlen(temp)+1);
  strcpy(r, temp);
  return(r);
}

char *alloc_item(s, num)
     char *s;
     int num;
{
  return(alloc_o_item(',', s, num));
}
   

/* CLASS_MATCH
 *
 * This routine compares a list found in a character string with a list found
 * in a LIST structure. If the string "all" is found in the character string,
 * then the function returns TRUE, or if either list has anything else in
 * common, TRUE is returned.
 */
int class_match(l, s)
     List l;
     char *s;
{
  int num, k;
  char item[256], *m_class;

  if (*s == 0)                                 /* Match all if null */
    return(TRUE);

  num = num_items(s);                          /* How many things to check */

  for(k=1; k <= num; k++) {
    get_item(s, item, k);                      /* Get one item */

    if (!strcmp(item, "all"))                  /* Is it 'all' */
      return(TRUE);
    
    SetBeginning(l);                           /* See if this exists in List */
    while(!EndList(l)) {
      ReadData(l, 0, (void **) &m_class);
      if (!strcmp(item, m_class))
	return(TRUE);
      SetNext(l);
    }
  }
  return(FALSE);                               /* Could not find a match */
}
 
/* MY_GETS
 *
 * This routine simulates the fgets routine, except that if the first character
 * in a string is '#', denoting a comment, the line is skipped.
 */
char *my_gets(s, length, fp)
     char *s;
     int length;
     FILE *fp;
{
  char *t;

  do {
    t = fgets(s, length, fp);
    *(s+strlen(s)-1) = '\0';
    if (t == 0 || (*s != '#' && *s != '\0'))
      return(t);
  } while(TRUE);
}

char *SaveString(s)
     char *s;
{
  char *t;

  t = (char *) malloc(strlen(s)+1);
  strcpy(t, s);
  return(t);
}


double get_seconds()
{
  double secs;

  struct timeval tp;
  struct timezone tzp;

  gettimeofday(&tp, &tzp);
  secs = (double) tp.tv_sec + (double) ((double)tp.tv_usec)/(double)1000000.0;
  return(secs);
}
  
