#include <varargs.h>
#include <netdb.h>
#include <sys/types.h>
#include <netinet/in.h>

#include "sockets.h"
#include "spede.h"
#include "objects.h"


int _type_size[] = {0, 4, 2, 4, 20, 10};


void pack_link_des(link)
     link_des *link;
{
  link->node.addr = htonl(link->node.addr);
  link->obj_ref = htonl(link->obj_ref);
}


void *pack_object(object)
     object_instance *object;
{
  object_instance *temp;
  
  temp = (object_instance *) malloc(sizeof(object_instance));
  *temp = *object;

  pack_link_des(&temp->creator);
  pack_link_des(&temp->storesite);
  pack_link_des(&temp->location);
  temp->id = htonl(object->id);
  
  return((void *) temp);
}


void unpack_link_des(link)
     link_des *link;
{
  link->node.addr = ntohl(link->node.addr);
  link->obj_ref = ntohl(link->obj_ref);
}


void unpack_object(data, object)
     void *data;
     object_instance *object;
{
  *object = *(object_instance *) data;
  unpack_link_des(&object->creator);
  unpack_link_des(&object->storesite);
  unpack_link_des(&object->location);

  object->id = ntohl(object->id);
}



object_instance *new_object()
{
  object_instance *new;

  new = (object_instance *) malloc(sizeof(object_instance));
  new->type[0] = 0;
  new->id = 0L;
  new->is_fault_tolerant = new->debug = 0;

  return(new);
}

void free_object(obj)
     object_instance *obj;
{
  free(obj);
}


parameter_block *new_params(va_alist)
     va_dcl
{
  va_list ap;
  register int k;
  int num_params, count = 0;
  long size;
  parameter_block *p;
  
  p = (parameter_block *) malloc(sizeof(parameter_block));
  for(k=0; k < MAX_PARAMS; k++) {
    p->param[k].data = 0;
    p->param[k].len = 0;
  }
  
  va_start(ap);
  num_params = va_arg(ap, int);
  p->num_params = num_params;
  
  while(num_params) {
    size = va_arg(ap, int);
    p->param[count].data = (void *) malloc(size);
    p->param[count].len = size;
    count++;
    num_params--;
  }
  
  va_end(ap);
  return(p);
}


void set_params(va_alist)
     va_dcl
{
  va_list ap;
  parameter_block *p;
  void *data;
  int num_params = 0;

  va_start(ap);
  p = va_arg(ap, parameter_block *);
  if (p == 0)
    return;

  while((p->num_params > num_params)) {
    if (data = va_arg(ap, void *))
      memcpy(p->param[num_params].data, data, p->param[num_params].len);
    num_params++;
  }
  va_end(ap);
}


void get_params(va_alist)
     va_dcl
{
  va_list ap;
  parameter_block *p;
  char **data;
  int num_params = 0;

  va_start(ap);
  p = va_arg(ap, parameter_block *);
  if (p == 0)
    return;

  while((p->num_params > num_params)) {
    if (data = va_arg(ap, char **))
      *data = p->param[num_params].data;
    num_params++;
  }
  va_end(ap);
}



void get_param_len(va_alist)
     va_dcl
{
  va_list ap;
  parameter_block *p;
  int *len;
  int num_params = 0;

  va_start(ap);
  p = va_arg(ap, parameter_block *);
  if (p == 0)
    return;

  while((p->num_params > num_params)) {
    if (len = va_arg(ap, int *))
      *len = p->param[num_params].len;
    num_params++;
  }
  va_end(ap);
}



void clean_params(p)
     pblock *p;
{
  int k;
  
  if (p == 0)
    return;
  for (k=0; k < p->num_params; k++)
    free(p->param[k].data);
}

void free_params(p)
     pblock *p;
{
  if (p == 0)
    return;
  clean_params(p);
  free(p);
}


int type_size(type)
     int type;
{
  return(_type_size[type]);
}


return_info *new_rinfo()
{
  return_info *new;

  new = (return_info *) malloc(sizeof(return_info));
  return(new);
}

