/* RWSock.c                 Jim Gochee                11/17/90
 *
 * This library can be used to send information across a given port number
 * which is passed to all of the routines. The information is read/written in
 * chunks, such as integers, longs, strings, and other formats. Use these
 * routines in two applications which are going to talk using these calls.
 *
 */

#define DEFAULT_WAIT 25                 /* Number of seconds before timeout */

#include <sys/time.h>
#include <stdio.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <math.h>

#include "select.h"
#include "rw_sock.h"

#define DEFAULT_BUFF_SIZE 20000
#define EXTRA 100;
extern int errno;
char *outbuff = 0;
char *read_buff;
int read_type = SOCKET;
long buff_bytes = 0;
long buff_size = 0;



long get_num_bytes()
{
  return(buff_bytes);
}


/* _WRITE_DATA
 *
 * This is the low level write routine that actually puts data onto the wire
 */
int _write_data(fd, buff, len)
     int fd;
     char *buff;
     int len;
{
  fd_set local_fd_connect;
  static struct timeval ti;
  int nfound, c=0, num;
  
  ti.tv_sec = DEFAULT_WAIT;   /* How long to wait while trying */
  ti.tv_usec = 0;
  
  FD_ZERO(&local_fd_connect);
  FD_SET(fd, &local_fd_connect);

  while(1) {
    nfound = select(FD_SETSIZE, 0, &local_fd_connect, 0, &ti);
    if (nfound != 1) {
      if (c == 2)
	return(SOCK_WRITE_ERR);
      c++;
      continue;
    }
    break;
  }    
  
  num = write(fd, buff, len);
  if (num == -1) {
    exit(1);
  }

  if (num != len) {
    exit(0);
  }

  return(0);
}



/* FLUSH_BUFF
 *
 * This routine actually sends the data
 */
int flush_buff(fd)
     int fd;
{
  int error;
  
  error = _write_data(fd, outbuff, buff_bytes);
  buff_bytes = 0;
  return(error);
}



/* WRITE_BLOCK
 *
 * This routine is responsible for putting data into the outbuff
 */
void write_block(buff, len)
     void *buff;
     long len;
{

  if (len < 1)
    return;

  if (outbuff == 0) {
    outbuff = (char *) malloc(DEFAULT_BUFF_SIZE);
    buff_size = DEFAULT_BUFF_SIZE;
  }

  if (len + buff_bytes > buff_size) {
    buff_size = buff_bytes + len + EXTRA;
    outbuff = (char *) realloc((char *) outbuff, buff_size);
/*    printf("Increasing buff_size to %d\n", buff_size);*/
  }

/*  printf("  Inserting %d bytes at position %d\n", len, buff_bytes); */
  bcopy(buff, outbuff+buff_bytes, len);
  buff_bytes += len;
}


void write_char(_c)
     int _c;
{
  char c;
  
  c = _c;
  write_block(&c, 1);
}


void write_header(_h)
     int _h;
{
  char h;

  h = _h;
  write_block(&h, 1);
}


void write_short(i)
     int i;
{
  u_short netshort;
  
  netshort = htons(i);
  write_block(&netshort, 2);
}


void write_long(l)
     long l;
{
  u_long netlong;

  netlong = htonl(l);
  write_block(&netlong, 4);
}


void write_string(s)
     char *s;
{

  if (s == 0) {
    write_short(0);
    return;
  }

  write_short(strlen(s));
  write_block(s, strlen(s));
}


void write_double(d)
     double d;
{
  char temps[30];

  sprintf(temps, "%.18g", d);
  write_string(temps);
}


void set_read(set, buff)
     int set;
     char *buff;
{
  read_type = set;
  read_buff = buff;
}
    


int read_chunk(socket, buff, max)
     int socket;
     char *buff;
     long max;
{
  int num_read;

  num_read = read(socket, buff, max);
  if (num_read == 0)
    num_read = -1;
  return(num_read);
}


int read_block(fd, buff, num)
     int fd;
     void *buff;
     long num;
{
  int error, numbytes;
  static fd_set *local_fd_connect = 0L;
  static struct timeval ti;
  int nfound, total_rec;
  static FILE *fp=0;
  
  if (num < 1)
    return(0);
  
  switch(read_type) {
  case BUFFER:
    bcopy(read_buff, buff, num);
    read_buff += num;
    return(0);
  case SOCKET:
    ti.tv_sec = DEFAULT_WAIT;   /* Wait for a certain time period seconds */
    ti.tv_usec = 0;
    
    if (local_fd_connect == 0L) {
      local_fd_connect = (fd_set *) malloc(sizeof(fd_set));
    }
    
    FD_ZERO(local_fd_connect);
    FD_SET(fd, local_fd_connect);
    
    total_rec = 0;
    while(total_rec < num) {
      nfound = select(FD_SETSIZE, local_fd_connect, 0, 0, &ti);
      if (nfound == -1)
	return(SOCK_READ_ERR);
      
      if (nfound == 0) {
	return(SOCK_READ_ERR);
      }
      
      numbytes = read(fd, ((char *)buff) + total_rec, num-total_rec);
      if (numbytes < 1)
	return(SOCK_READ_ERR);
      
      total_rec += numbytes;
    }
    return(0);
  } /* switch */
  return(0);
}


int read_char(fd, c)
     int fd;
     char *c;
{
  return(read_block(fd, c, 1));
}


int read_header(fd, h)
     int fd;
     char *h;
{
  return(read_block(fd, h, 1));
}


int read_short(fd, i)
     int fd;
     short *i;
{
  int error;
  u_short netshort;

  if (error = read_block(fd, &netshort, 2))
    return(error);

  *i = ntohs(netshort);
  return(0);
}


int read_long(fd, l)
     int fd;
     long *l;
{
  int error;
  u_long netlong;

  if (error = read_block(fd, &netlong, 4))
      return(error);
  
  *l = ntohl(netlong);
  return(0);
}


int read_string(fd, s, max)
     int fd;
     char *s;
     int max;
{
  int error = 0;
  short size;
  char c;

  error |= read_short(fd, &size);
  if (size < 1) {
    *s = 0;
    return(error);
  }
  
  max--;  /* Make sure we have room for the end '\0' */
  if (size >= max) {
    if (error = read_block(fd, s, max))
      return(error);
    size -= max;  /* How many characters left over */
    while (size > 0) {
      if (error = read_char(fd, &c))
	return(error);
      size --;
    }
    s[max] = '\0';
  }
  else {
    if (error = read_block(fd, s, size))
      return(error);
    s[size] = '\0';
  }
  return(0);
}



int read_a_string(fd, s)
     int fd;
     char **s;
{
  int error;
  short size;
  char c;

  error = read_short(fd, &size);

  *s = (char *) malloc(size+1);
  
  error |= read_block(fd, *s, size);
  *(*s + size) = 0;

  return(error);
}


int read_double(fd, d)
     int fd;
     double *d;
{
  int error;
  char temp[255];

  if (error = read_string(fd, temp, 255))
    return(error);

  *d = (double) atof(temp);
  return(0);
}
  
