#include <stdio.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>

#include "select.h"
#include "sockets.h"


#define SLEEP_INTERVAL 2                /* How long to sleep between tries */
#define TRUE 1
#define FALSE 0

extern int errno;


void hent_to_ndes(h, n)
     struct hostent *h;
     node_des *n;
{
  if (!h)
    return;

  strcpy(n->name, h->h_name);
  n->addr = *(long *) h->h_addr;
}


/* NEW_HOSTENT
 *
 * Allocates memory for a new host entry.
 */
struct hostent *new_hostent()
{
  register struct hostent *h;

  h = (struct hostent *) malloc(sizeof(struct hostent));
  h->h_name = 0;
  h->h_aliases = 0;
  h->h_addrtype = 0;
  h->h_length = 0;
  h->h_addr_list = 0;

  return(h);
}


/* FREE_HOSTENT
 *
 * Deallocates memory for a host entry
 */

void free_hostent(h)
     struct hostent *h;
{
  register int i;

  if (h->h_name)
    free(h->h_name);
  
  if (h->h_aliases) {
    i = 0;
    while(h->h_aliases[i]) {
      free(h->h_aliases[i]);
      i++;
    }
    free(h->h_aliases);
  }

  if (h->h_addr_list) {
    free(h->h_addr_list[0]);
    free(h->h_addr_list);
  }
}


/* INSTALL_LISTENER
 *
 * This routine installs an internet socket at the given port and returns 
 * the fileno of the socket. If a socket cannot be opened at that port, 
 * the value zero is returned and the global 'errno' contains the error.
 */
int install_listener(the_port, num_connections)
     int the_port;
     int num_connections;
{
  int sock;
  struct sockaddr_in me;

  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)
    return(-1);

  me.sin_family = AF_INET;
  me.sin_addr.s_addr = INADDR_ANY;
  me.sin_port = htons(the_port);

  if (bind(sock, &me, sizeof(me))) {
    close(sock);
    return(-1);
  }

  listen(sock, num_connections);
  
  return(sock);
}


void random_listener(the_sock, port, connects)
     int *the_sock;
     int *port;
     int connects;
{
  int sock;
  struct sockaddr_in me;
  long length;

  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    *the_sock = -1;
    return;
  }

  me.sin_family = AF_INET;
  me.sin_addr.s_addr = INADDR_ANY;
  me.sin_port = 0;

  if (bind(sock, &me, sizeof(me))) {
    *the_sock = -1;
    close(sock);
    return;
  }

  length = sizeof(me);
  getsockname(sock, &me, &length);

  listen(sock, connects);
  *the_sock = sock;
  *port = ntohs(me.sin_port);
}
  
  

/* COPY_NODEDES
 *
 */
void copy_node_des(dest, src)
     node_des *dest;
     node_des *src;
{
  *dest = *src;
}



/* COPY_HOSTENT
 *
 * This routine copies and allocates space in the destination hostent.
 */
void copy_hostent(dest, src)
     struct hostent *dest, *src;
{
  dest->h_name = (char *) malloc(strlen(src->h_name) + 1);
  strcpy(dest->h_name, src->h_name);
  dest->h_aliases = 0;
  dest->h_addrtype = src->h_addrtype;
  dest->h_length = src->h_length;
  dest->h_addr_list = (char **) malloc(2*src->h_length);
  dest->h_addr_list[1] = 0;
  dest->h_addr_list[0] = (char *) malloc(src->h_length);
  bcopy(src->h_addr_list[0], dest->h_addr_list[0], src->h_length);
}

/* MY_GETHOSTBYADDR
 *
 * This routine fills in a hostent struct with the info returned from a 
 * gethostbyaddr call.
 */
int my_gethostbyaddr(addr, h)
     long addr;
     struct hostent *h;
{
  struct hostent *hp;

  if ((hp = gethostbyaddr(&addr, 4, AF_INET)) == 0) {
    h->h_name = 0;
    return(-1);                                          /* Failure */
  }
  
  copy_hostent(h, hp);
  return(0);
}


/* MY_GETNODEBYADDR
 *
 * This routine fills in a hostent struct with the info returned from a 
 * gethostbyaddr call.
 */
int my_getnodebyaddr(addr, node)
     long addr;
     node_des *node;
{
  struct hostent *hp;

  if ((hp = gethostbyaddr(&addr, 4, AF_INET)) == 0) {
    return(-1);                                          /* Failure */
  }

  strcpy(node->name, hp->h_name);
  node->addr = addr;
  return(0);
}


long get_remote_addr(machine)
     char *machine;
{
  struct hostent *hp;

  hp = gethostbyname(machine);
  if (hp == 0)
    return(0);

  return(*(long *) hp->h_addr);
}


/* GET_LOCAL_ADDRESS
 *
 * This routine fills in the address parameter to reflect the local machine.
 * It returns zero if successfull, or an error if not successfull.
 */
int get_local_address(address)
     long *address;
{
  char hostname[256];
  struct hostent *hp;
  
  gethostname(hostname, 255);
  if ((hp = gethostbyname(hostname)) == 0)
    return(errno);

  bcopy(hp->h_addr, address, 4);
  return(0);
}


/* ACCEPT_CONNECTION
 *
 * This routine accepts a connection on a given socket and fills in the
 * calling machine's info. If successfull, the routine returns a socket, else 
 * a zero with 'errno' containing the error code.
 */
int accept_connection(the_socket, remote_node)
     int the_socket;
     node_des *remote_node;
{
  int new_sock;
  struct sockaddr_in connecting;
  int length;
  struct hostent *hp;
  
  length = sizeof(connecting);
  new_sock = accept(the_socket, &connecting, &length);
  if (new_sock == -1) {
    printf("Error %d accepting a new connection.\n", errno);
  }
  
  hp = gethostbyaddr(&connecting.sin_addr, 4, AF_INET);
  if (hp == 0) {
    fprintf(stderr, "Warning!!! Unknown connection from addr %X on socket %d\n", connecting.sin_addr, new_sock);
    return(new_sock);
  }

  hent_to_ndes(hp, remote_node);

  return(new_sock);
}


/* CONNECT_REMOTE
 *
 * This routine establishes a connection with a address/port on a remote
 * machine. If successfull, the routine returns the socket of the connection,
 * otherwise 0 is returned.
 *
 * PARAMETERS:
 *            port - the remote port number
 *            addr - the remote address
 *         timeout - the number of seconds to try (0 for a single try)
 */
int connect_remote(port, addr, timeout)
     int port, timeout;
     long addr;
{
  int sock, connected;
  struct sockaddr_in remote;
  struct timeval the_time;
  struct timezone tz;
  int end_time;
  
  gettimeofday(&the_time, &tz);                      /* Used for the timeout */
  end_time = the_time.tv_sec + timeout;
  
  connected = FALSE;
  while(!connected) {
    if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0)    /* Build the socket */
      return(-1);

    remote.sin_family = AF_INET;
    remote.sin_addr.s_addr = addr;
    remote.sin_port = htons(port);
    if (connect(sock, &remote, sizeof(remote))) {
      if (errno != 61)                                /* Fatal connect error */
	return(-1);

      close(sock);
      gettimeofday(&the_time, &tz);
      if (the_time.tv_sec >= end_time)        /* Has timeout been exceeded? */
	return(-1);                           /* Yes, so return a bad status */
      sleep(SLEEP_INTERVAL);
    }
    else
      connected = TRUE;
  }

  return(sock);
}


int connect_remote_h(host, port, timeout)
     char *host;
     int port;
     int timeout;
{
  struct hostent *hp;

  hp = gethostbyname(host);
  if (hp == 0)
    return(-1);
  return(connect_remote(port, *(long *) hp->h_addr, timeout));
}


int socket_has_data(socket)
     int socket;
{
  fd_set fd_connect;
  struct timeval ti;

  ti.tv_sec = 0;
  ti.tv_usec = 0;
  FD_ZERO(&fd_connect);
  FD_SET(socket, &fd_connect);
  
  return(select(FD_SETSIZE, &fd_connect, 0, 0, &ti));
}
