#include <stdio.h>
#include <sys/types.h>                    /* This is necessary on northstar */
#include <netinet/in.h>                   /* This is necessary on northstar */

/* These include files are necessary for objects */
#include "sockets.h"
#include "spede.h"
#include "spede_io.h"
#include "objects.h"
#include "objlib.h"

#include "misc.h"

char *object_name = "test_obj";          /* This is used to register obj */
double setup_time = 0.0;

#define TEST_MESSAGE 11

#define TRUE 1
#define FALSE 0

/* CALLBACK
 *
 * This is the main callback routine for the object. When an object receives
 * an OBJ_MESSAGE command, this routine will be called to handle the message.
 * The message number is passed in as 'message', parameters to the object are
 * passed in in 'params', and 'r_params' is used to send parameters back to
 * the caller. The underlying mechanism will transport the return paramters and
 * free any memory associated with them when the transfer is complete.
 *
 * The routine also returns a long 'result' value, which is sent to the calling
 * object. Note that it is not necessary to have any return parameters in
 * 'r_params'.
 *
 * The two messages that all objects should pay attention to are "CREATE" and
 * "DELETE". These are predefined constants. CREATE is sent to an object when
 * it is first created, and DELETE is sent just before the object is
 * terminated. NOTE: User message numbers must be greater than 10.
 */
static long callback(message, params, r_params)
     int message;
     parameter_block *params, **r_params;
{
  long result;

  r_params = 0L;
  switch(message) {
  case CREATE:                                 /* Allows object to alloc mem */
    break;
  case DELETE:                                 /* Tell object to shut down */
    break;
  case TEST_MESSAGE:                           /* Don't do anything */
    return(1);
    break;
/* case user_messages:
   break;
*/
  default:
    printf("Invalid message.\n");
    return(0);
  }
}
  

/* MAIN
 *
 * This main calls two underlying routines for object operation. The user need
 * not put any code in this area. If the object in question is launched from
 * the command line, the routine 'free_launch' will be called.
 */
main(argc, argv) 
     int argc;
     char **argv;
{
  setup_time = get_seconds();
  if(init_objects(argc, argv, callback)) {
    perror("Error trying to initialize the object library");
    exit(1);
  }
  register_dean();                      /* Tell local dean we exist */
}


object_instance **create_objects(obj_type, num_objects)
     char *obj_type;
     int *num_objects;
{
  char *machine;
  int k, m, err, num_create, num_machines;
  object_instance **objs;
  link_des link;
  char **the_machines;
  extern object_descriptor run_info;

  printf("Creating the work objects\n");

  /* Query the 'reg' daemon for machines of the given class */
  num_machines = run_info.num_machines;

  num_create = run_info.num_objects;
  
  printf("Creating from a list of %d machines\n", num_machines);
  printf("Attempting to creating %d objects\n", num_create);

  objs = (object_instance **) malloc(sizeof(object_instance *) * num_create);

  m = 0;
  for(k=0; k < num_create;) {
    if (num_machines < 1) {
      printf("There are no machines specified!!\n");
      exit(0);
    }
    machine = run_info.machine[m % num_machines];
    printf("Asynchronously creating object %d on machine '%s'\n", k, machine);
    objs[k] = create_object(machine, obj_type, FALSE, FALSE, TRUE);
    if (objs[k] != 0) {
      k++;
    }
    else {
      printf("Failed creating on that machine...trying another.\n");
      run_info.machine[m] = run_info.machine[num_machines-1];
      num_machines--;
    }
    m++;
  }

  err = sync_created_objects();
  *num_objects = k;
  return(objs);
}


void object_done(object, rinfo, num_running)
     object_instance *object;
     return_info  *rinfo;
     int *num_running;
{
  (*num_running)--;
}


/* FREE_LAUNCH
 * 
 * This routine is called by the object library if the user launched the
 * object from the command line. Some objects should not be launched from the
 * command line and can terminate if this routine is called. Other objects will
 * use routine as if it were 'main'.
 */
void free_launch(argc, argv)
     int argc;
     char **argv;
{
  static int num_running;
  object_instance **the_objects;
  int num_objects, num_remote = 1, k, j;
  double start_time, end_time;

  the_objects = create_objects("test_obj", &num_objects);
  printf("There are %d test objects, setup time %.4g secs\n", num_objects, get_seconds() - setup_time);

  printf("Timing %d remote invokations for %d objects\n", num_remote, num_objects);
  start_time = get_seconds();
  
  for(k=0; k < num_remote; k++) {
    num_running = num_objects;
    for(j=0; j < num_objects; j++) {
      send_obj_mess_callback(the_objects[j], TEST_MESSAGE, 0, object_done, &num_running);
    }
    while(num_running)
      idle_objects(20, 0);
  }
  
  end_time = get_seconds();
  printf("Total time was %.4g seconds\n", num_remote, end_time-start_time);
}


/* NORMAL_LAUNCH
 *
 * This routine is the counterpart to free_launch. It is called by the object
 * library when the object is launched in response to a 'create_object'
 * command from another object. This routine is given as another place to
 * put initialization code.
 */

void normal_launch()
{
}
