/* REG.C
 * Jim Gochee
 * 5/17/92
 *
 * This program acts as a Registering service for entities communicating in the
 * SPEDE environment. This program also allows the user (from a command line)
 * to launch new deans and stop running deans.
 */

#include <stdio.h>
#include <sys/time.h>
#include <sys/errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <varargs.h>
#include <utmp.h>
#include <sys/file.h>
#include <ctype.h>

#include "sockets.h"
#include "select.h"
#include "network.h"
#include "timed_task.h"
#include "misc.h"
#include "spede.h"
#include "spede_io.h"

#define TRUE 1
#define FALSE 0

#define USER_DELIM   '#'               /* What the user delims classes with */
#define CON_DELIM    '|'               /* What reg.machines delims with     */
#define LAUNCH_TIMEOUT      17         /* How long before launch timeout */


typedef struct _machine_ent {
  char name[33];
  char class[100];
  char base_dir[100];
  Nodeptr nodeptr;
  long addr;
} machine_ent;


typedef struct _connect_ent {
  machine_ent *the_machine;
  char service[33];
  long port_num;
  int socket;
  Nodeptr nodeptr;
} connected_ent;


extern int errno;

int world_connect;
long my_addr;
connect_type *connections;
node_des local_node;                 /* This host's info                     */
List connect_list;
List machine_list;                   /* List of known machines */
List class_list;                     /* List of known classes */
List launch_list;                    /* List of launch requests */
int line;
timedtasks ts;
long tickle_id;
int port_num;
char hostname[32], *HOME_PATH;

#define SHOW   1
#define CONFIG 2
#define START  3
#define STOP   4
#define TICKLE 5
#define QUIT   6
#define MACHINES 7
#define HELP1  8
#define HELP2  9

char *command_list[] = {"show","config","start","stop","tickle","quit","machines", "help", "?", 0};
char *config_commands[] = {"classes","machines",0};


/* PARSE_COMMAND
 *
 * This routine parses a user command into one of the commands defined in
 * a command list. If no command within the command list of specified, the
 * routine will return 0. Otherwise it will return the command index + 1
 */
int parse_command(s, commands)
     char *s;
     char **commands;
{
  char comm[100];
  char *t, old;
  int k;

  t = s;
  while(*t != 0 && *t != ' ')
    t++;
  old = *t;
  *t = 0;
  strcpy(comm, s);
  *t = old;

  k=0;
  while(commands[k] != 0) {
    if (!strcmp(comm, commands[k]))
      return(k+1);
    k++;
  }
  return(0);
}


/* MPRINTF_RESET
 *
 * Resets the number of lines that have been printed.
 */
void mprintf_reset()
{
  line = 0;
}


/* VFPRINTF
 *
 * This hack is necessary for some machines that do not have vfprintf
 * defined. Note that only arguements 4 bytes in size will work. 
 */
#ifndef vfprintf
vfprintf(stream,fmt,args)
FILE *stream;
char *fmt, *args[];
{
  return(fprintf(stream,fmt,args[0],args[1],args[2],args[3],args[4],args[5]));
}
#endif


/* MPRINTF
 *
 * "MORE-PRINTF" This function was designed to allow the user to see a screen
 * full of data with a pause. The function keeps track of how many lines have
 * been printed so far, and when the limit is reached, the user is asked for
 * input.
 */
void mprintf(format, va_alist)
     char *format;
     va_dcl
{
  va_list ap;
  char ch;

  if (line > 20) {
    line = 0;
    printf("--Hit Return--");
    fflush(stdout);
    ch = getchar();
  }

  line++;
  va_start(ap);
  vfprintf(stdout, format, ap);
  va_end(ap);
}


/* COMPARE_CLASS
 *
 * This routine will compare a class as specified by a user, with a class
 * in the form of the config file. The function returns true if the user's
 * class list is equivalent to the config class list.
 */
int compare_class(u_class, class_list)
     char *u_class;
     char *class_list;
{
  int num_classes, num_test, k, j, found;
  char a_class[33], test_class[33];
  
  num_classes = num_o_items(USER_DELIM, u_class);
  num_test = num_o_items(CON_DELIM, class_list);

  for(k=1; k <= num_classes; k++) {
    get_o_item(USER_DELIM, u_class, a_class, k);
    found = FALSE;
    for(j=1; j <= num_test; j++) {
      get_o_item(CON_DELIM, class_list, test_class, j);
      if (!strcmp(test_class, a_class)) {
	return(TRUE);
      }
    }
  }
  return(FALSE);
}

    
/* VALID_CLASS
 *
 * A valid class is defined as a class descriptor, which can include multiple
 * classes, where all of the classes mentioned must have been registered at the
 * head of the config file.
 */
int valid_class(class)
     char *class;
{
  char *s;
  int num_classes, k;
  char a_class[33];

  if (!strcmp(class, "*"))                        /* Wildcard Class */
    return(TRUE);

  num_classes = num_o_items(USER_DELIM, class);
  for(k=1; k <= num_classes; k++) {
    get_o_item(USER_DELIM, class, a_class, k);
    SetBeginning(class_list);
    while(!EndList(class_list)) {
      ReadData(class_list, (void **) &s, 0);
      if (!strcmp(a_class, s))
	break;
      SetNext(class_list);
    }
    if (EndList(class_list))            /* Did we run out of things to check */
      return(FALSE);
  }
  return(TRUE);
}



/* INIT_VALID_CLASS
 *
 * A valid class is defined as a class descriptor, which can include multiple
 * classes, where all of the classes mentioned must have been registered at the
 * head of the config file.
 */
int init_valid_class(class)
     char *class;
{
  char *s;
  int num_classes, k;
  char a_class[33];

  num_classes = num_o_items(CON_DELIM, class);
  for(k=1; k <= num_classes; k++) {
    get_o_item(CON_DELIM, class, a_class, k);
    SetBeginning(class_list);
    while(!EndList(class_list)) {
      ReadData(class_list, (void **) &s, 0);
      if (!strcmp(a_class, s))
	break;
      SetNext(class_list);
    }
    if (EndList(class_list))            /* Did we run out of things to check */
      return(FALSE);
  }
  return(TRUE);
}


/* INSERT_CLASS
 *
 * This function inserts all machines matching the given class into the
 * list 'dest'.
 */
int insert_class(class, src, dest)
     char *class;
     List src, dest;
{
  machine_ent *ent;
  int wildcard, num_classes, k;
  void *temp;

  if (!valid_class(class))
    return(FALSE);

  wildcard = (!strcmp(class, "*"));

  SetBeginning(src);
  while(!EndList(src)) {
    ReadData(src, &temp, (void **) &ent);
    if (compare_class(class, ent->class) || wildcard)
      InsertData(dest, temp, (void *) ent);
    SetNext(src);
  }
  return(TRUE);
}


/* INSERT_MACHINE
 *
 * This routine inserts all machines matching the given machine name into
 * the list 'dest'.
 */
int insert_machine(s, src, dest)
     char *s;
     List src, dest;
{
  machine_ent *ent;
  int wildcard, num=0;
  void *temp;

  wildcard = (!strcmp(s, "*"));

  SetBeginning(src);
  while(!EndList(src)) {
    ReadData(src, &temp, (void **) &ent);
    if (!strcmp(ent->name, s) || wildcard) {
      InsertData(dest, temp, (void *) ent);
      num++;
    }
    SetNext(src);
  }
  return(num);
}
      

/* CONSTRUCT_SCOPE
 *
 * This routine takes a scope parameter, which is a string describing either
 * a list of machines or a list of classes, and builds a list of all the
 * matching machines.
 */
void construct_scope(src, dest, s)
     List src, dest;
     char *s;
{
  int num_params, k;
  char *t, *r;

  num_params = num_o_items(' ', s);
  for(k=1; k <= num_params; k++) {  
    t = alloc_o_item(' ', s, k);
    if (*t == '[') {                     /* Brackets denote a class list */
      t++;
      if ((r = strchr(t, ']')))
	*r = 0;
      if (!insert_class(t, src, dest))
	printf("%% Invalid class name '%s'\n", t);
      t--;
    }
    else {
      if (!insert_machine(t, src, dest))
	printf("%% Invalid machine name '%s'\n", t);
    }
    free(t);
  }
}
    
    
/* STOP_SERVICE
 *
 * This routine is called when the user wants to stop an executing service.
 */
void stop_service(s)
     char *s;
{
  connected_ent *ent;
  command comm;
  char service[33];
  char *t;
  List scope_list;
  machine_ent *me;

  if (num_o_items(' ', s) < 3) {
    printf(" Format: stop <service> <scope1> [scope2],... (<scope> = <machine | [class]>)\n\n");
    return;
  }

  get_o_item(' ', s, service, 2);
  scope_list = CreateList();
  t = ptr_of_o_item(' ', s, 3);
  construct_scope(connect_list, scope_list, t);
  
  comm.comm = KILL;
  SetBeginning(scope_list);
  while(!EndList(scope_list)) {
    ReadData(scope_list, (void **) &ent, 0);
    if (!strcmp(ent->service, service) || *service == '*') {
    printf("Sending terminate to '%s' at socket %d\n", ent->the_machine->name, ent->socket);
      write_command(&comm, ent->socket);
  }
    SetNext(scope_list);
  }    
  FreeList(scope_list);  
}


/* _START_SERVICE
 *
 * This is the low level routine to start a service on a remote host. The
 * service is started with the 'system' command by using 'rsh' to launch
 * the correct service on the remote machine.
 */
void _start_service(service, machine, base_dir)
     char *service, *machine, *base_dir;
{
  char tempstr[256];
  char comm[256];
  
  printf("\nStarting service '%s' on machine '%s'\n", service, machine);

  sprintf(comm, "rsh %s -n \"%s",machine, base_dir);
  if (base_dir[0] != 0)                      /* Is there a base dir? */
    strcat(comm, "/");                       /* Yes, so need a backslash */
  
  strcat(comm, service);
  sprintf(tempstr, " -h %s -p %d", hostname, port_num);
  strcat(comm, tempstr);
	  
  strcat(comm, ">& /dev/null&\" > /dev/null 2>&1&");

  printf("  ->"); puts(comm);
  
  system(comm);
}

  
/* START_SERVICE
 *
 * This routine is responsible for interpreting the command line for launching
 * remote services. Services can be launch on more machines than one.
 */
void start_service(s)
     char *s;
{
  char service[33];
  char *t;
  List scope_list;
  machine_ent *me;

  if (num_o_items(' ', s) < 3) {
    printf(" Format: start <service> <scope1> [scope2],... (<scope> = <machine | [class]>)\n\n");
    return;
  }

  get_o_item(' ', s, service, 2);
  scope_list = CreateList();
  t = ptr_of_o_item(' ', s, 3);
  construct_scope(machine_list, scope_list, t);
  
  SetBeginning(scope_list);
  while(!EndList(scope_list)) {
    ReadData(scope_list, 0, (void **) &me);
    _start_service(service, me->name, me->base_dir);
    SetNext(scope_list);
  }
  FreeList(scope_list);
}
  
  
/* PRINT_HEADER
 *
 * Print the command prompt
 */
void print_header()
{
  printf("Command: ");
  fflush(stdout);

}


/* PRINT_CONNECTED
 *
 * Print machines that are currently connect with reg.
 */
int print_connected()
{
  connected_ent *ent;
  int num=0;
  

  mprintf_reset();

  mprintf("\n");
  mprintf("Machine                           Port   Socket  Type\n");
  mprintf("=====================================================\n");
  SetBeginning(connect_list);
  while(!EndList(connect_list)) {
    num++;
    ReadData(connect_list, (void **) &ent, 0);
    mprintf("%-32s  %-5d  %-5d  %s\n", ent->the_machine->name, ent->port_num, ent->socket, ent->service);
    SetNext(connect_list);
  }
  if (!num)
    mprintf("(No machines connected)\n\n");
  else
    mprintf("\n%d machines connected\n\n", num);
}


/* F_TICKLE_FUNC
 *
 * This function is a callback for when tickle is on and NO_OP packets must
 * be sent to all connected services.
 */
int f_tickle_func(data)
     void *data;
{
  connected_ent *ent;
  command comm;

  build_NO_OP(&comm);
  SetBeginning(connect_list);
  while(!EndList(connect_list)) {
    ReadData(connect_list, (void **) &ent, 0);
    write_command(&comm, ent->socket);
    SetNext(connect_list);
  }
  return(0);
}


/* TICKLE_SERVICES
 *
 * This function controls whether or not the user is tickling all services.
 */
void tickle_services(s)
     char *s;
{
  char on_off[10];

  get_o_item(' ', s, on_off, 2);
  if (!strcmp(on_off, "off")) {
    printf("Tickle turned off\n");
    remove_timed_task(ts, tickle_id);
  }
  else {
    printf("Tickle turned on\n");
    f_tickle_func(0);                        /* Do a tickle now */
    tickle_id = add_timed_task(ts, 60, 0, f_tickle_func, (void *) 0);
  }
}


/* SHOW_MACHINES
 *
 * This routine shows all machines, regardless of whether or not they are
 * connect with reg.
 */
void show_machines(s)
     char *s;
{
  machine_ent *ent;

  mprintf_reset();

  mprintf("\n");
  mprintf("Machine                           Address  Classes\n");
  mprintf("=======================================================\n");
  SetBeginning(machine_list);
  while(!EndList(machine_list)) {
    ReadData(machine_list, 0, (void **) &ent);
    mprintf("%-32s %X  %s\n", ent->name, ent->addr, ent->class);
    SetNext(machine_list);
  }
}
    

/* QUIT
 *
 * This function is called when the user quits from reg. It attemps to shut
 * down all open sockets, without specifically killing each entity. This allows
 * deans to remain running, which is desirable because they will re-connect
 * with the next reg that runs.
 */
quit()
{
  connected_ent *ent;
  
  SetBeginning(connect_list);
  while(!EndList(connect_list)) {
    ReadData(connect_list, (void **) &ent, 0);
    DeleteNode(connect_list);
    shutdown(ent->socket, 2);
    close(ent->socket);
    free(ent);
  }

  shutdown(ent->socket, 2);
  close(world_connect);
}


/* SHOW_HELP
 * 
 * This functions prints out recognized commands
 */
void show_help()
{
  char s[256];

  printf("\n\n");
  printf("List of commands:\n");
  printf("  show                      - Prints all connected objects\n");
  printf("  machines                  - Prints all known machine entries\n");
  printf("  start dean <machine info> - Starts a dean on given machines\n");
  printf("  stop dean <machine info>  - Stops deans on given machines\n");
  printf("  tickle <on | off>         - Keep deans alive?\n");
  printf("  quit                      - Exit the program\n");
  printf("\n\n");
  printf("Hit return:");
  gets(s);
}


/* HANDLE_STDIN
 * 
 * This function will read and parse input from the user (command line).
 */
void handle_stdin(socket)
     int socket;
{
  char s[256];
  int num;
  
  gets(s);

  num = parse_command(s, command_list);
  switch(num) {
  case SHOW:
    print_connected();
    break;
  case CONFIG:
    break;
  case START:
    start_service(s);
    break;
  case STOP:
    stop_service(s);
    break;
  case TICKLE:
    tickle_services(s);
    break;
  case MACHINES:
    show_machines(s);
    break;
  case HELP1:
  case HELP2:
    show_help();
    break;
  case QUIT:
    quit();
    exit(0);
  }

  print_header();
}


/* MACHINE_FROM_NAME
 *
 * This routine returns a machine_ent type from a name. Since unix machines
 * can have various names, the addresses of the machines are used. 
 */
machine_ent *machine_from_name(s)
     char *s;
{
  long r_addr;
  machine_ent *the_ent;

  r_addr = get_remote_addr(s);

  SetBeginning(machine_list);
  while(!EndList(machine_list)) {
    ReadData(machine_list, 0, (void **) &the_ent);
    if (the_ent->addr == r_addr)
      return(the_ent);
    SetNext(machine_list);
  }
  return(0);
}

/* NEW_SERV
 *
 * This function is called when a new service has been launched. All entities
 * in the SPEDE network will register themselves with reg. In doing this, they
 * send a special "register service" packet which has the service name.
 */
void new_serv(comm, ce)
     command *comm;
     connect_entry *ce;
{
  connected_ent *ent;
  machine_ent *me;
  command *temp_c;
  command r_comm;
  long r_addr;
  int task;

  ent = (connected_ent *) malloc(sizeof(connected_ent));

  ent->socket = ce->the_socket;
  ent->the_machine = me = machine_from_name(comm->data.reg_serv.machine);
  if (ent->the_machine == 0) {
    printf("\nNew service '%s' registered on unknown machine '%s'\n", comm->data.reg_serv.type, comm->data.reg_serv.machine);
    return;
  }

  strcpy(ent->service, comm->data.reg_serv.type);
  ent->port_num = comm->data.reg_serv.port_num;
  ce->data = ent;

  SetEnd(connect_list);
  ent->nodeptr = InsertData(connect_list, (void *) ent, (void *) me);
  printf("\nNew service '%s' launched on machine '%s' through sock %d.\n", ent->service, ent->the_machine->name, ce->the_socket);
  print_header();

  /* Send the service its machine base directory. Base directories tell
   * services where to find and put certain files.
   */

  build_BASE_DIR(&r_comm, ent->the_machine->base_dir);
  write_command(&r_comm, comm->socket);

  SetBeginning(launch_list);
  while(!EndList(launch_list)) {
    ReadData(launch_list, (void **) &task, (void **) &temp_c);
    if (!temp_c) {
      SetNext(launch_list);
      continue;
    }

    r_addr = get_remote_addr(temp_c->data.l_serv.machine);

    if ((r_addr == ent->the_machine->addr) && (!strcmp(temp_c->data.l_serv.service, ent->service))) {
      build_SERV_INFO(&r_comm, temp_c->message_id, ent->the_machine->name,
		      ent->service, ent->port_num, TRUE);
      write_command(&r_comm, temp_c->socket);
      free_command(temp_c);
      WriteData(launch_list, (void *) task, 0);
      return;
    }
    SetNext(launch_list);
  }
} 


/* NUMBER_CLASS
 *
 * This routine returns the number of machines in a list of classes.
 */
int number_class(class)
     char *class;
{
  machine_ent *the_ent;
  int count = 0;

  SetBeginning(machine_list);
  while(!EndList(machine_list)) {
    ReadData(machine_list, 0, (void **) &the_ent);
    if (compare_class(class, the_ent->class))
      count++;
    SetNext(machine_list);
  }
  return(count);
}


/* MACHINE_NAMES
 *
 * This routine returns an array of machine names which match the class
 * specifyer.
 */
char **machine_names(class, num)
     char *class;
     int num;
{
  char **s, *temp;
  machine_ent *the_ent;
  int count = 0, tempcount, sw, k;
  
  s = (char **) malloc(num * sizeof(char *));
  SetBeginning(machine_list);
  while(!EndList(machine_list) && num) {
    ReadData(machine_list, 0, (void **) &the_ent);
    if (compare_class(class, the_ent->class)) {
      s[count++] = the_ent->name;
      num--;
    }
    SetNext(machine_list);
  }

  /*
   * Randomize the names
   */
  for(k=0; k < count; k++) {
    sw = (rand() % count);
    temp = s[k];
    s[k] = s[sw];
    s[sw] = temp;
  }

  return(s);
}


/* SERV_INFO
 *
 * This routine answers a request for service information. Typically, when
 * a free-launched object wants to connect to remote deans, it must ask the
 * reg for the remote dean's information, such as if the dean is actually
 * running and on what port.
 */
void serv_info(comm, ce)
     command *comm;
     connect_entry *ce;
{
  command s_comm;
  char **machines;
  int num;
  char *class = comm->data.class.class;

  switch(comm->data.class.control) {
  case 0:
    build_R_CLASS_INFO(&s_comm, comm->send_ref, comm->message_id, number_class(class), 0);
    write_command(&s_comm, comm->socket);
    break;
  case 1:
    num = number_class(class);
    printf("Asked for info about class '%s', which has %d elements\n", class, num);
    num = (num < comm->data.class.num) ? num : comm->data.class.num;
    build_R_CLASS_INFO(&s_comm, comm->message_id, comm->send_ref, num, (machines = machine_names(class, num)));
    write_command(&s_comm, comm->socket);
    free(machines);
    break;
  }
}


/* F_LAUNCH_TIMEOUT
 *
 * This routine is called after a specified number of seconds to see how a
 * given launch request is proceeding. If the request has been fullfilled, then
 * nothing is done. If the launch request has not been filled, then reg
 * assumes that the requested service cannot be run on the remote machine and
 * the orinally caller is notified. 
 */
int f_launch_timeout(data)
     void *data;
{
  Nodeptr nodeptr = (Nodeptr) data;
  command comm, *temp_c;
  int task;

  SetNode(launch_list, nodeptr);
  ReadData(launch_list, (void **) &task, (void **) &temp_c);
  if (temp_c != 0) {
    printf("LAUNCH TIMEOUT for machine '%s'\n", temp_c->data.l_serv.machine);
    build_SERV_INFO(&comm, temp_c->message_id, temp_c->data.l_serv.machine, temp_c->data.l_serv.service, 0, FALSE);
    write_command(&comm, temp_c->socket);
    free_command(temp_c);
  }
  remove_timed_task(ts, task);
  DeleteNode(launch_list);
  return(0);
}
  

/* LAUNCH_SERVICE
 *
 * This functions takes a command from the net to launch a remote service and
 * processes it.
 */
void launch_service(c, ce)
     command *c;
     connect_entry *ce;
{
  machine_ent *me;
  long r_addr;
  command *temp_c;
  Nodeptr nodeptr;
  int task_id;

  r_addr = get_remote_addr(c->data.l_serv.machine);
  if (r_addr == 0)
    printf("Cannot get remote address of machine '%s'\n", c->data.l_serv.machine);

  SetBeginning(machine_list);
  while(!EndList(machine_list)) {
    ReadData(machine_list, 0, (void ** )&me);
    if (me->addr == r_addr) {
      _start_service(c->data.l_serv.service, me->name, me->base_dir);
      SetEnd(launch_list);
      temp_c = new_command();
      *temp_c = *c;
      nodeptr = InsertData(launch_list, 0, 0);
      task_id = add_timed_task(ts, LAUNCH_TIMEOUT, 0, f_launch_timeout, nodeptr);
      SetNode(launch_list, nodeptr);
      WriteData(launch_list, (void *) task_id, temp_c);
      return;
    }
    SetNext(machine_list);
  }
  printf("ERROR!!! Cannot launch machine '%s'. It is not in my machine list\n", c->data.l_serv.machine);
  
  
}
 

/* F_QUERY_SERV
 *
 * This function returns information about a service.
 */
void f_query_serv(c, ce)
     command *c;
     connect_entry *ce;
       
{
  connected_ent *cc;
  machine_ent *me;
  long r_addr;
  command comm;
  long found_port = -1;
  int alive;

  r_addr = get_remote_addr(c->data.query_serv.machine);

  SetBeginning(connect_list);
  while(!EndList(connect_list)) {
    ReadData(connect_list, (void **) &cc, (void **) &me);
    if (r_addr == me->addr && 
	(!strcmp(cc->service, c->data.query_serv.service))) {
      found_port = cc->port_num;
      break;
    }
    SetNext(connect_list);
  }

  if (found_port != -1)
    build_SERV_INFO(&comm, c->message_id, c->data.query_serv.machine,
		    c->data.query_serv.service, found_port, TRUE);
  else
    build_SERV_INFO(&comm, c->message_id, c->data.query_serv.machine,
		    c->data.query_serv.service, -1, FALSE);
  write_command(&comm, c->socket);
}


/* REG_COMMAND
 *
 * This routine parses an incoming network command and calls an appropriate
 * function. 
 */
int reg_command(comm, ce)
     command *comm;
     connect_entry *ce;
{
  int err;

  switch(comm->comm) {
  case REGISTER_SERV:
    new_serv(comm, ce);
    break;
  case GET_CLASS_INFO:
    serv_info(comm, ce);
    break;
  case LAUNCH_SERVICE:
    launch_service(comm, ce);
    break;
  case QUERY_SERV:
    f_query_serv(comm, ce);
    break;
  default:
    printf("Unknown command ($%X) on socket %d\n", comm->comm, ce->the_socket);
    break;
  }
  return(0);
}


/* SOCKET_WENT_DOWN
 * 
 * This function is called when a socket connection is lost
 */
int socket_went_down(ce)
     connect_entry *ce;
{
  connected_ent *ent, *c_ent;

  close_commands(ce->the_socket);

  ent = ce->data;
  printf("\nConnection with '%s' running '%s' on sock %d  went down\n", ent->the_machine->name, ent->service, ce->the_socket);

  SetNode(connect_list, ent->nodeptr);
  ReadData(connect_list, (void **) &c_ent, 0);

  DeleteNode(connect_list);
  free(ent);
  delete_connection(connections, ce);
  print_header();
}
  

/* MAIN_LOOP
 *
 * This function loops through all open sockets and stdin, looking for input
 * of some sort. If data is found, appropriate routines are called.
 */
int main_loop()
{
  int err;
  int status;
  struct timeval ti, notify_time;
  struct timezone tz;
  int nfound, new_socket, socket, test_connect, result;
  connect_entry *ce;
  long output_time, check_time;
  long check_id;
  int new_sock;
  command comm;
  
  connections = new_connections();
  ce = add_connection(connections, world_connect, &local_node);
  ce = add_connection(connections, fileno(stdin), &local_node);
  ts = new_timed_tasks();

  printf("Command: ");
  fflush(stdout);

  while(TRUE) {
    next_task_wait(ts, &ti); 
    nfound = wait_on_connections(connections, &ti);    

    /* Loop for the number of sockets with data */
    if (nfound) {
      while(ce = get_next_entry(connections)) {
	if (ce->the_socket == world_connect) {
	  ce = accept_new_connection(connections, world_connect);
	  open_commands(ce->the_socket);
	}
	else if (ce->the_socket == fileno(stdin)) {
	  handle_stdin(ce->the_socket);
	}
	else {                                 /* Must be a command */
	  socket = ce->the_socket;
	  if ((status = complete_command(&comm, ce->the_socket))) {
	    if (status == -1)
	      socket_went_down(ce);
	    else
	      reg_command(&comm, ce);
	  }
	}
      }
    }
    result = do_tasks(ts);
    if (result) {
      printf("REG - No user interaction in a while... exiting\n");
      break;
    }
  }

  free_timed_tasks(ts);
  return(result);
}

/* ADD_CLASS
 *
 * This routine adds a class to the list of known classes.
 */
void add_class(s, line)
     char *s;
     int line;
{
  char *t;

  t = (char *) malloc(strlen(s)+1);
  strcpy(t, s);

  SetEnd(class_list);
  InsertData(class_list, t, 0);
}


/* ADD_MACHINE
 *
 * This routine adds a machine to the list of known machines.
 */
void add_machine(s, line)
     char *s;
     int line;
{
  machine_ent *ent;

  if (num_items(s) != 3) {
    printf("%% Illegal number of items for a machine description in line %d\n", line);
    exit(0);
  }

  ent = (machine_ent *) malloc(sizeof(machine_ent));
  get_item(s, ent->name, 1);
  get_item(s, ent->class, 2);
  if (!init_valid_class(ent->class)) {
    printf("%% Illegal class descriptor '%s' on line %d\n", ent->class, line);
    exit(0);
  }

  get_item(s, ent->base_dir, 3);
  ent->addr = get_remote_addr(ent->name);
  if (ent->addr == 0) {
    printf("Warning...Could not get address of machine '%s'. Discarding that machine.\n", ent->name);
    free(ent);
  }
  else {
    SetEnd(machine_list);
    ent->nodeptr = InsertData(machine_list, 0, (void *) ent);
  }
}

  
/* LOAD_CONFIG
 * 
 * This routine loads the machine list file.
 */
void load_config()
{
  FILE *fp;
  char s[256], *t;
  int mode=0;
  int line=0;

  sprintf(s, "%s/.reg.machines", HOME_PATH);
  fp = fopen(s,"r");
  if (fp == 0) {
    printf("Warning...There is no ~/.reg.machines\n");
    return;
  }

  while(fgets(s, 256, fp) != NULL) {
    if (s[0] != 0)
      *(s + strlen(s)-1) = 0;                 /* Remove ending '\n' char */
    line++;
    t = s;

    if (*t == '#')
      continue;
    if (*t == '%') {
      t++;                                    /* Get past the % */
      mode = parse_command(t, config_commands);
      if (mode == 0) {
	printf("%% Illegal config command on line %d\n", line);
	exit(0);
      }
      continue;
    }

    /* Kill white space */
    while (*t != 0 && isspace(*t))
      t++;

    if (*t == 0)
      continue;

    switch(mode) {
    case 1:
      add_class(s, line);
      break;
    case 2:
      add_machine(s, line);
      break;
    default:
      printf("%% Found data without a command on line %d\n", line);
      exit(0);
    }
  }
  fclose(fp);
} 


/* READ_INFO_FILE
 *
 * This routine reads the reg info file in the user's home directory. The
 * info file records the last reg host that executed. 
 */
void read_info_file(old_host, old_port)
     char *old_host;
     int *old_port;
{
  FILE *fp;
  char fname[256];

  sprintf(fname, "%s/.reginfo", HOME_PATH);
  fp = fopen(fname, "r");
  if (fp == 0) {
    *old_port = 0;
    old_host[0] = '\0';
    return;
  }

  fgets(old_host, 32, fp);
  old_host[strlen(old_host)-1] = '\0';
  fscanf(fp, "%d", old_port);
  fclose(fp);
}


/* SET_UP_LISTENER
 *
 * This routine sets up the main reg socket listener. It chooses the port
 * to use by first seeing if it can reconnect to the last port used by a 
 * reg. This is done in case deans are still executing. If the old port
 * cannot be used, then a random one is used.
 */
void set_up_listener(old_port)
     int old_port;
{
  int choice;

  while(TRUE) {
    /* Try to launch from the same old port num */
    world_connect = install_listener(old_port, 20);
    if (world_connect < 0) {
      printf("Failed to open connection on old port number %d\n", old_port);
      printf("Do you want to: (0)-Try again, (1)-Use a new port: ");
      choice = 1;
      scanf("%d", &choice);
      if (choice == 0)
	continue;
      else {
	random_listener(&world_connect, &port_num, 20);
	return;
      }
    }
    else {
      port_num = old_port;
      return;
    }
  }
}


/* MAIN
 *
 */
main(argc, argv)
     int argc;
     char **argv;
{
  int err;
  char fname[256];
  struct linger my_linger;
  FILE *fp;
  char our_port[10];
  char our_machine[256];
  char old_host[256];
  int old_port;

  HOME_PATH = (char *) getenv("HOME");

  read_info_file(old_host, &old_port);
  printf("Reg last ran on %s, port %d\n", old_host, old_port);
  gethostname(hostname, 32);

  if (!strcmp(old_host, hostname)) {
    set_up_listener(old_port);
  }
  else
    random_listener(&world_connect, &port_num, 20);
  
  sprintf(fname, "%s/.reginfo", HOME_PATH);
  fp = fopen(fname, "w");
  fputs(hostname, fp);
  fprintf(fp, "\n%d\n", port_num);
  fclose(fp);

  my_linger.l_onoff = 0;
  setsockopt(world_connect, SOL_SOCKET, SO_LINGER, &my_linger, sizeof(my_linger));

  if (world_connect == -1) {
    fprintf(stderr, "Error #%d trying to install a listener at port %d\n", errno, port_num);
    exit(1);
  }
  
  get_local_address(&my_addr);
  my_getnodebyaddr(my_addr, &local_node);
  
  printf("\nREG - Running on %s (%X), port %d\n", hostname, ntohl(my_addr), port_num); 

  connect_list = CreateList();
  class_list = CreateList();
  machine_list = CreateList();
  launch_list = CreateList();

  load_config();
  main_loop();
}
